require("prefabutil")

local assets =
{
    Asset("ANIM", "anim/chandelier_archives.zip"),
    Asset("ANIM", "anim/chandelier_fire.zip"),
}

local prefabs =
{
    "chandelier_fire",
}

local ON = 1
local OFF = 2

local light_params =
{
    on =
    {
        radius = 5,
        intensity = .6,
        falloff = .6,
        colour = { 131/255, 194/255, 255/255 },
        time = 3,
    },

    off =
    {
        radius = 0,
        intensity = 0,
        falloff = 1,
        colour = { 0, 0, 0 },
        time = 3,
    },
}

local FLAMEDATA = {
    "flame1",
    "flame2",
    "flame3",
    "flame4",
}

local light_phases = {}
for k, v in pairs(light_params) do
    table.insert(light_phases, k)
    v.id = #light_phases
    v.tint = { v.colour[1] * .5, v.colour[2] * .5, v.colour[3] * .5, 0--[[ alpha, zero for additive blending ]] }
end

local function pushparams(inst, params)
    inst.Light:SetRadius(params.radius * inst.widthscale)
    inst.Light:SetIntensity(params.intensity)
    inst.Light:SetFalloff(params.falloff)
    inst.Light:SetColour(unpack(params.colour))

    if params.intensity > 0 then
        inst.Light:Enable(true)
    else
        inst.Light:Enable(false)			
	end

end

-- Not using deepcopy because we want to copy in place
local function copyparams(dest, src)
    for k, v in pairs(src) do
        if type(v) == "table" then
            dest[k] = dest[k] or {}
            copyparams(dest[k], v)
        else
            dest[k] = v
        end
    end
end

local function lerpparams(pout, pstart, pend, lerpk)
    for k, v in pairs(pend) do
        if type(v) == "table" then
            lerpparams(pout[k], pstart[k], v, lerpk)
        else
            pout[k] = pstart[k] * (1 - lerpk) + v * lerpk
        end
    end
end

local function OnUpdateLight(inst, dt)
    inst._currentlight.time = inst._currentlight.time + dt
    if inst._currentlight.time >= inst._endlight.time then
        inst._currentlight.time = inst._endlight.time
        inst._lighttask:Cancel()
        inst._lighttask = nil
    end

    lerpparams(inst._currentlight, inst._startlight, inst._endlight, inst._endlight.time > 0 and inst._currentlight.time / inst._endlight.time or 1)
    pushparams(inst, inst._currentlight)
    inst.AnimState:SetLightOverride(Remap(inst._currentlight.intensity, light_params.off.intensity,light_params.on.intensity, 0,1))
    for k, v in pairs(FLAMEDATA) do
        if inst[v] then
            local val = Remap(inst._currentlight.intensity, light_params.off.intensity,light_params.on.intensity, 0,1)
            inst[v].AnimState:SetLightOverride(val)
            inst[v].Transform:SetScale(val,val,val)
            if inst._currentlight.intensity == 0 and val == 0 then
                inst[v]:Remove()
                inst[v] = nil
            end
        else
            if inst._currentlight.intensity > 0 then
                local fx = SpawnPrefab("chandelier_fire")
                --inst:AddChild(fx)
                fx.entity:AddFollower()
                fx.Follower:FollowSymbol(inst.GUID, v,0,0,0)
                inst[v] = fx
                fx.Transform:SetScale(0,0,0)
            end
        end
    end
end

local function OnLightPhaseDirty(inst)
    local phase = light_phases[inst._lightphase]
    if phase ~= nil then
        local params = light_params[phase]
        if params ~= nil and params ~= inst._endlight then
            copyparams(inst._startlight, inst._currentlight)
            inst._currentlight.time = 0
            inst._startlight.time = 0
            inst._endlight = params
            if inst._lighttask == nil then
                inst._lighttask = inst:DoPeriodicTask(FRAMES, OnUpdateLight, nil, FRAMES)
            end
        end
    end
end

local function updatelight(inst)
    if not inst:HasTag("on_ground_now") then
        if inst.components.playerprox:IsPlayerClose() and inst._lightphase ~= ON then
	        inst._lightphase = ON
            OnLightPhaseDirty(inst)
    	elseif inst._lightphase ~= OFF then
	        inst._lightphase = OFF
            OnLightPhaseDirty(inst)		
        end
	end
end

local function onhammered(inst)
    inst.components.lootdropper:DropLoot()
    local fx = SpawnPrefab("collapse_big")
    fx.Transform:SetPosition(inst.Transform:GetWorldPosition())
	fx.SoundEmitter:PlaySound("dontstarve/common/destroy_metal")
    inst:Remove()
end

local function onhit(inst)
    inst.AnimState:PlayAnimation("hit")
    inst.AnimState:PushAnimation("ground_idle")
end

local function GetStatus(inst)
    return inst:HasTag("on_ground_now")
        and "ONGROUND"
        or nil
end

local function BecomeWorkable(inst, workleft)	
    inst:AddComponent("workable")
    inst.components.workable:SetWorkAction(ACTIONS.HAMMER)
    inst.components.workable:SetWorkLeft(workleft or 6)
    inst.components.workable:SetOnFinishCallback(onhammered)
    inst.components.workable:SetOnWorkCallback(onhit)	
end

local NON_SMASHABLE_TAGS = { "INLIMBO", "irreplaceable" }
local function Smash(inst)	
	local x, y, z = inst.Transform:GetWorldPosition()
    local ents = TheSim:FindEntities(x, 0, z, 3, nil, NON_SMASHABLE_TAGS)

	for k,v in pairs(ents) do
	   	if v ~= nil and v ~= inst then  
		    if v.components.combat ~= nil then
	    		v.components.combat:GetAttacked(inst, 100, nil)
			elseif v.components.workable ~= nil then
			    v.components.workable:WorkedBy(inst, 20)
			end
	   	end
	end	
end

local function IsPointNearNonLilyPadPuddle(pt, radius)
    local puddle = TheSim:FindEntities(pt.x, pt.y, pt.z, radius, {"puddle"}, {"puddle_disappear"})
    local lilypad = TheSim:FindEntities(pt.x, pt.y, pt.z, radius, {"LilyPadForPuddle"})	
    if #puddle > 0 and #lilypad <= 0 then
	    return true
	else
	    return false
	end
end

local function Fall(inst, playfallanim, workleft)
    if playfallanim then
        inst.AnimState:PlayAnimation("fall")     
	   	
	    local x, y, z = inst.Transform:GetWorldPosition()
	    if (inst.IsPosSurroundedByLand ~= nil and not inst:IsPosSurroundedByLand(x, y, z, 1.5)) or IsPointNearNonLilyPadPuddle(inst:GetPosition(), 2.5) then
		    inst.persists = false
		    inst:ListenForEvent("animover", function() 
			    if IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
			        SpawnPrefab("splash_water").Transform:SetPosition(x, y, z)
				end
				for k, v in pairs(FLAMEDATA) do
				    if inst[v] then
                        inst[v]:Remove()
                        inst[v] = nil
					end
				end
				inst.components.lootdropper:DropLoot()
		        inst:Remove()
			end)
		else
		    inst.AnimState:PushAnimation("ground_idle")
	        inst:DoTaskInTime(0.5, Smash)
     	end		
		
    else
	    inst.AnimState:PlayAnimation("ground_idle")     
	end
    inst:AddTag("on_ground_now") 
    inst:RemoveTag("NOBLOCK")
    inst:RemoveTag("ancient_hanging_lamp")
    inst:RemoveTag("fallable")
	BecomeWorkable(inst, workleft)
    if inst._lightphase ~= OFF then
	    inst._lightphase = OFF
        OnLightPhaseDirty(inst)		
    end
	for k, v in pairs(FLAMEDATA) do
    	if inst[v] then
            inst[v]:Remove()
            inst[v] = nil
		end
	end		
end

local function OnSave(inst, data)
    data.on_ground_now = inst:HasTag("on_ground_now") or nil
	data.workleft = inst.components.workable ~= nil and inst.components.workable.workleft or nil
end

local function OnLoad(inst, data)
    if data ~= nil then
	    if data.on_ground_now then
            inst:Fall(nil, data.workleft)
	    end
    end	
end

local function fn()
    local inst = CreateEntity()

    inst.entity:AddTransform()
    inst.entity:AddAnimState()
    inst.entity:AddLight()
    inst.entity:AddSoundEmitter()
    inst.entity:AddMiniMapEntity()

    inst.MiniMapEntity:SetIcon("ancient_hanging_lamp_item.tex")

    inst.AnimState:SetBloomEffectHandle("shaders/anim.ksh")
    inst.AnimState:SetBank("chandelier_archives")
    inst.AnimState:SetBuild("chandelier_archives")
    inst.AnimState:PlayAnimation("idle", true)
	inst.AnimState:SetRayTestOnBB(true)

    inst:AddTag("NOBLOCK")
    inst:AddTag("ancient_hanging_lamp")
    inst:AddTag("fallable")
	inst:AddTag("absorb_immune")
	inst:AddTag("structure")

    inst.widthscale = 1
    inst._endlight = light_params.off
    inst._startlight = {}
    inst._currentlight = {}
    copyparams(inst._startlight, inst._endlight)
    copyparams(inst._currentlight, inst._endlight)
    pushparams(inst, inst._currentlight)

    inst._lightphase = OFF
    inst._lighttask = nil

    inst:AddComponent("inspectable")
	inst.components.inspectable.getstatus = GetStatus

    inst:AddComponent("lootdropper")
	inst.components.lootdropper:SetLoot({"nightmarefuel","nightmarefuel","nightmarefuel","thulecite","thulecite","bluegem"})

    inst:AddComponent("playerprox")
    inst.components.playerprox:SetDist(20, 23) --15,17
    inst.components.playerprox:SetOnPlayerNear(updatelight)
    inst.components.playerprox:SetOnPlayerFar(updatelight)
	
	inst.Fall = Fall
	
	inst.OnSave = OnSave
	inst.OnLoad = OnLoad
	
	return inst
end

local function firefxfn()
    local inst = CreateEntity()

    inst.entity:AddTransform()
    inst.entity:AddAnimState()

    inst.AnimState:SetBloomEffectHandle("shaders/anim.ksh")
    inst.AnimState:SetBank("chandelier_fire")
    inst.AnimState:SetBuild("chandelier_fire")
    inst.AnimState:PlayAnimation("idle", true)
	inst.AnimState:SetSortOrder(4)		

    inst:AddTag("NOBLOCK")
    inst:AddTag("NOCLICK")
    inst:AddTag("FX")	
	
	inst.persists = false
	
	return inst
end

local function OnHitWater(inst)
	if inst.float_fx == nil then
		inst.float_fx = SpawnPrefab("float_fx_front")
		inst.float_fx.entity:SetParent(inst.entity)
		inst.float_fx.AnimState:PlayAnimation("idle_front_large", true)		
	end
	if inst.float_fx2 == nil then
		inst.float_fx2 = SpawnPrefab("float_fx_back")
		inst.float_fx2.entity:SetParent(inst.entity)
		inst.float_fx2.AnimState:PlayAnimation("idle_back_large", true)
	end	
end

local function OnHitLand(inst)
	if inst.float_fx ~= nil then
		inst.float_fx:Remove()
		inst.float_fx = nil
	end
	if inst.float_fx2 ~= nil then
		inst.float_fx2:Remove()
		inst.float_fx2 = nil
	end	
end

local function ondeploy(inst, pt, deployer)
    local ancient_hanging_lamp = SpawnPrefab("ancient_hanging_lamp") 
    if ancient_hanging_lamp then 
        ancient_hanging_lamp.Transform:SetPosition(pt.x, pt.y, pt.z)
        ancient_hanging_lamp.SoundEmitter:PlaySound("dontstarve/common/place_structure_stone")
        inst:Remove()
    end 
end

local function IsPointNearAncientHangingLamp(pt, range)
    for i, v in ipairs( TheSim:FindEntities(pt.x, 0, pt.z, range, {"ancient_hanging_lamp"}) ) do
        if v:GetDistanceSqToPoint(pt) < range * range then
            return true
        end
    end
    return false
end
--[[
local function IsPointOutsideInteriorWall(pt, radius)
    if TheCamera.interior then
        local interiorSpawner = GetWorld().components.interiorspawner 
        if GetWorld() and interiorSpawner and interiorSpawner.current_interior then
            local originpt = interiorSpawner:getSpawnOrigin()
            local width = interiorSpawner.current_interior.width
            local depth = interiorSpawner.current_interior.depth        
            local dMax = originpt.x + depth/2
            local dMin = originpt.x - depth/2
            local wMax = originpt.z + width/2
            local wMin = originpt.z - width/2        
            local dist = radius or 0.5
            if pt.x < dMin+dist or pt.x > dMax-dist or pt.z < wMin+dist or pt.z > wMax-dist then
                return true
            end
        end
    end
    return false
end
]]
local function deploytest(inst, pt)
	local radius = 2	
    for i = -radius, radius, 1 do
        if GetWorld().Map:GetTileAtPoint(pt.x - radius, pt.y, pt.z + i) == GROUND.IMPASSABLE or GetWorld().Map:GetTileAtPoint(pt.x + radius, pt.y, pt.z + i) == GROUND.IMPASSABLE then
            return false
        end
    end
	
    for i = -(radius - 1), radius - 1, 1 do
        if GetWorld().Map:GetTileAtPoint(pt.x + i, pt.y, pt.z - radius) == GROUND.IMPASSABLE or GetWorld().Map:GetTileAtPoint(pt.x + i, pt.y, pt.z + radius) == GROUND.IMPASSABLE then
            return false
        end
    end			
	
	return not IsPointNearAncientHangingLamp(pt, 6) and not TheCamera.interior --and not IsPointOutsideInteriorWall(pt, 3)  --TO DO: Make the indoor lamp targetable
	                                                --Crash when entering a interior with ancient_hanging_lamp...
end

local function itemfn()
    local inst = CreateEntity()

    inst.entity:AddTransform()
    inst.entity:AddAnimState()

    inst.AnimState:SetBank("chandelier_archives")
    inst.AnimState:SetBuild("chandelier_archives")
    inst.AnimState:PlayAnimation("ground_idle")
    inst.AnimState:SetScale(0.75, 0.75, 0.75)	

    MakeInventoryPhysics(inst)

    inst:AddTag("eyeturret")  --For action string
    inst:AddTag("suitable_for_puddle")
 
    inst:AddComponent("inspectable")

    inst:AddComponent("inventoryitem")

    if inst.GetIsOnWater ~= nil then
        inst:AddComponent("floatable")
	    inst.components.floatable:SetOnHitWaterFn(OnHitWater)
    	inst.components.floatable:SetOnHitLandFn(OnHitLand)
	    inst.components.inventoryitem:SetOnPutInInventoryFn(OnHitLand)
    end	

    inst:AddComponent("deployable")
    inst.components.deployable.ondeploy = ondeploy
	inst.components.deployable.test = deploytest
    inst.components.deployable.placer = "ancient_hanging_lamp_placer"
    inst.components.deployable.deploydistance = 1.5

	return inst
end

return Prefab("ancient_hanging_lamp", fn, assets, prefabs),
    Prefab("chandelier_fire", firefxfn, assets),
	MakePlacer("ancient_hanging_lamp_placer", "chandelier_archives", "chandelier_archives", "idle"),
	Prefab("ancient_hanging_lamp_item", itemfn, assets)
